#include <cstdio>
#include <cstdlib>
#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  pragma comment(lib, "glew32.lib")
#  include "glew.h"
#  include "glut.h"
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

/*
** 点の数
*/
#define POINTS 100000

/*
**トラックボール処理
*/
#include "Trackball.h"
static Trackball trackball;
static int pressedButton;

/*
** 変換行列
*/
#include "Matrix.h"
static Matrix projectionMatrix;
static Matrix modelviewMatrix;
static GLint transformMatrixLocation;

/*
** シェーダ
*/
#include "shadersource.h"
static const char vertSource[] = "simple.vert";
static const char fragSource[] = "simple.frag";
static GLuint gl2Program;

/*
** 図形
*/
#include "disseminate.h"
static GLuint buffer[1];
static GLsizei points;
static GLint pointLocation;

/*
** 画面表示
*/
static void display(void)
{
  /* 投影変換／視野変換／モデリング変換 */
  Matrix transformMatrix = projectionMatrix * modelviewMatrix * trackball.get();
  transformMatrix.scale(1.0f, 1.0f, 2.0f);
  transformMatrix.translate(0.0f, 0.0f, -0.5f);
  
  /* シェーダプログラムを適用する */
  glUseProgram(gl2Program);

  /* uniform 変数 transformMatrix に行列を設定する */
  glUniformMatrix4fv(transformMatrixLocation, 1, GL_FALSE, transformMatrix.get());

  /* attribute 変数 point に頂点情報を対応付ける */
  glEnableVertexAttribArray(pointLocation);

  /* 頂点バッファオブジェクトとして buffer[0] を指定する */
  glBindBuffer(GL_ARRAY_BUFFER, buffer[0]);

  /* 頂点情報の格納場所と書式を指定する */
  glVertexAttribPointer(pointLocation, 3, GL_FLOAT, GL_FALSE, 0, 0);

  /* 画面クリア */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* 隠面消去処理を有効にする */
  glEnable(GL_DEPTH_TEST);

  /* 図形を描く */
  glDrawArrays(GL_POINTS, 0, points);

  /* 隠面消去処理を無効にする */
  glDisable(GL_DEPTH_TEST);

  /* 頂点バッファオブジェクトを解放する */
  glBindBuffer(GL_ARRAY_BUFFER, 0);

  /* index が 0 の attribute 変数の頂点情報との対応付けを解除する */
  glDisableVertexAttribArray(pointLocation);

  glutSwapBuffers();
}

/*
** ウィンドウサイズ
*/
static void resize(int w, int h)
{
  /* ビューポート */
  glViewport(0, 0, w, h);

  /* 透視投影変換行列を求める */
  projectionMatrix.loadIdentity();
  projectionMatrix.camera(30.0f, (GLfloat)w / (GLfloat)h, 5.0f, 9.0f);
  
  /* トラックボールする範囲 */
  trackball.region(w, h);
}

/*
** キーボード
*/
static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
    case '\033':
    case 'q':
    case 'Q':
      exit(0);
    default:
      break;
  }
}

/*
** アニメーション
*/
static void idle(void)
{
  /* 画面の描き替え */
  glutPostRedisplay();
}

/*
** マウスのクリック
*/
static void mouse(int button, int state, int x, int y)
{
  pressedButton = button;
  
  switch (pressedButton) {
    case GLUT_LEFT_BUTTON:
      if (state == GLUT_DOWN) {
        /* トラックボール開始 */
        trackball.start(x, y);
      }
      else {
        /* トラックボール停止 */
        trackball.stop(x, y);
      }
      break;
    default:
      break;
  }
}

/*
** マウスのドラッグ
*/
static void motion(int x, int y)
{
  switch (pressedButton) {
    case GLUT_LEFT_BUTTON:
      /* トラックボール移動 */
      trackball.motion(x, y);
      break;
    default:
      break;
  }
}

/*
** 初期化
*/
static void init(void)
{
  /* シェーダプログラムのコンパイル／リンク結果を得る変数 */
  GLint compiled, linked;
  
#if defined(WIN32)
  /* GLEW の初期化 */
  GLenum err = glewInit();
  if (err != GLEW_OK) {
    fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
    exit(1);
  }
#endif
  
  /* シェーダオブジェクトの作成 */
  GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
  GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
  
  /* シェーダのソースプログラムの読み込み */
  if (readShaderSource(vertShader, vertSource)) exit(1);
  if (readShaderSource(fragShader, fragSource)) exit(1);
  
  /* バーテックスシェーダのソースプログラムのコンパイル */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(vertShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in vertex shader.\n");
    exit(1);
  }
  
  /* フラグメントシェーダのソースプログラムのコンパイル */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(fragShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in fragment shader.\n");
    exit(1);
  }
  
  /* プログラムオブジェクトの作成 */
  gl2Program = glCreateProgram();
  
  /* シェーダオブジェクトのシェーダプログラムへの登録 */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);
  
  /* シェーダオブジェクトの削除 */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);
  
  /* シェーダプログラムのリンク */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &linked);
  printProgramInfoLog(gl2Program);
  if (linked == GL_FALSE) {
    fprintf(stderr, "Link error.\n");
    exit(1);
  }
  
  /* uniform 変数 transformMatrix の場所を得る */
  transformMatrixLocation = glGetUniformLocation(gl2Program, "transformMatrix");
  
  /* attribute 変数 point の場所を得る */
  pointLocation = glGetAttribLocation(gl2Program, "point");
  
  /* 視野変換行列を求める */
  modelviewMatrix.loadIdentity();
  modelviewMatrix.lookat(0.0f, 0.0f, 7.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
  
  /* 頂点バッファオブジェクトを１つ作る */
  glGenBuffers(1, buffer);
  
  /* 図形をバッファオブジェクトに登録する */
  points = disseminate(POINTS, buffer);
  
  /* トラックボール処理の初期化 */
  trackball.initialize();
  
  /* 背景色 */
  glClearColor(0.0, 0.1, 0.3, 1.0);
}

/*
** メインプログラム
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutIdleFunc(idle);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();

  return 0;
}
